{% extends "base.html" %}

{% block content %}
{% if page.nb_url %}
<a href="{{ page.nb_url }}" title="Download Notebook" class="md-content__button md-icon jp-DownloadNB">
    {% include ".icons/material/download.svg" %}
</a>
<a href="https://nbviewer.ipython.org/urls/sxs.readthedocs.io/en/main{{ page.nb_url }}" title="Open in nbviewer" class="md-content__button md-icon jp-DownloadNB">
    {% include ".icons/material/download.svg" %}
</a>
{% endif %}

{{ super() }}

{% endblock content %}
